/*********************************************************************
 * binary.c - Create binary package.
 *********************************************************************
 * This file is part of the package paco
 * Copyright (C) 2005 David Rosal <david.rosal@upf.edu>
 * For more information visit http://paco.sourceforge.net
 *********************************************************************/

#include "gpaco.h"
#include <sys/stat.h>

static struct {
	pkg_t *pkg;
	guint8 prog;
	guint8 level;
} zip;


static void zip_level_change(GtkWidget *zip_level_combo, __UNUSED gpointer data)
{
	zip.level = gtk_combo_box_get_active(GTK_COMBO_BOX(zip_level_combo)) + 1;
}


static void zip_prog_change(GtkWidget *zip_prog_combo, gpointer name_label)
{
	gchar *text, *pkg_name_8;
	
	zip.prog = gtk_combo_box_get_active(GTK_COMBO_BOX(zip_prog_combo));
	
	pkg_name_8 = gp_filename_to_utf8(zip.pkg->name);
	
	text = g_strdup_printf("%s-paco.tar.%s", pkg_name_8, (zip.prog == GP_GZIP) ? "gz" : "bz2");
	gtk_label_set_text(GTK_LABEL(name_label), text);
	
	g_free(text);
	g_free(pkg_name_8);
}


static void bin_destroy()
{
	_bin_dialog = NULL;
}


static void bin_response(GtkDialog *dialog, gint response_id, gpointer dir_button)
{
	file_t *file;
	GtkWidget *msg_dialog = NULL;
	FILE *ftmp;
	gint fd;
	GSList *f;
	gchar *error = NULL, *dir, *dir_8, *tar_path = NULL, *zip_path = NULL;
	gchar *zip_path_8 = NULL, *msg = NULL, *cmd = NULL, *tmpfile;
	gboolean empty, retval;
	struct stat s;
	
	if (response_id != GTK_RESPONSE_OK) {
		gtk_widget_destroy(GTK_WIDGET(dialog));
		return;
	}

	dir = gtk_file_chooser_get_filename(GTK_FILE_CHOOSER(dir_button));
	g_return_if_fail(dir != NULL);
	dir_8 = gp_filename_to_utf8(dir);
	
	/* Check whether we have write permissions on the dest. directory */
	if (access(dir, W_OK) < 0) {
		error = g_strdup_printf("%s: %s", dir_8, g_strerror(errno));
		gp_error_dialog(error, GTK_WIDGET(dialog), NULL);
		goto Return;
	}
	
	tar_path = g_strdup_printf("%s/%s-paco.tar", dir, zip.pkg->name);
	zip_path = g_strdup_printf("%s.%s", tar_path, (zip.prog == GP_GZIP) ? "gz" : "bz2");
	zip_path_8 = gp_filename_to_utf8(zip_path);
	
	if (g_lstat(zip_path, &s) == 0) {
		msg = g_strdup_printf(_("Package '%s' already exists.\nDo you want to overwrite it ?"),
		  zip_path_8);
		if (gp_question_dialog(GTK_WIDGET(dialog), msg) != GTK_RESPONSE_YES)
			goto Return;
	}
	  
	g_return_if_fail((fd = g_file_open_tmp("gpacoXXXXXX", &tmpfile, NULL)) != -1);
	close(fd);
	g_return_if_fail((ftmp = g_fopen(tmpfile, "w")));

	for (f = zip.pkg->files; f; f = f->next) {
		file = (file_t *)(f->data);
		if (g_lstat(file->name, &s) == 0)
			g_fprintf(ftmp, "%s\n", file->name);
	}
	empty = (ftell(ftmp) == 0);
	fclose(ftmp);
	if (empty) {
		gp_error_dialog(_("Empty package"), GTK_WIDGET(dialog), NULL);
		goto Return;
	}
	
	if (zip.pkg->size_inst > MEGABYTE) {
		msg = g_strdup_printf(_("Creating %s..."), zip_path_8);
		msg_dialog = gp_progress_dialog(msg, GTK_WIDGET(dialog), NULL);
	}	
	
	cmd = g_strdup_printf("tar cf %s --files-from=%s", tar_path, tmpfile);
	retval = gp_spawn(&error, cmd);
	gp_unlink(tmpfile);
	REFRESH_GUI;
	if (!retval) {
		gp_error_dialog(_("Error while creating the package:"), _bin_dialog, error);
		goto Return;
	}
	g_free(cmd);
	cmd = g_strdup_printf("%s -%df %s", (zip.prog == GP_GZIP)? "gzip" : "bzip2", zip.level,
	  tar_path);
	if (!gp_spawn(&error, cmd))
		gp_error_dialog(_("Error while creating the package:"), _bin_dialog, error);
	
	Return:
	g_free(msg);
	g_free(cmd);	
	g_free(dir);
	g_free(dir_8);
	g_free(error);
	g_free(tar_path);
	g_free(zip_path);
	g_free(zip_path_8);
	
	if (msg_dialog)
		gtk_widget_destroy(msg_dialog);
}


void binary_pkg(pkg_t *pkg)
{
	GtkWidget *hbox, *table, *label, *dir_button, *name_label, *combo;
	GtkBox *vbox;
	gchar *title, number[2], *pacoball, *pkg_name_8;
	guint8 i;

	g_return_if_fail(_bin_dialog == NULL);
	g_return_if_fail(pkg != NULL);
	g_return_if_fail(pkg->name != NULL);
	
	zip.pkg = pkg;
	pkg_name_8 = gp_filename_to_utf8(pkg->name);
	title = g_strdup_printf(_("%s - binary package"), pkg_name_8);
	
	_bin_dialog = gtk_dialog_new_with_buttons(
	   title, GTK_WINDOW(_window.window), GTK_DIALOG_DESTROY_WITH_PARENT,
	   GTK_STOCK_OK, GTK_RESPONSE_OK, GTK_STOCK_CANCEL, GTK_RESPONSE_CANCEL, NULL);
	gtk_window_set_resizable(GTK_WINDOW(_bin_dialog), FALSE);
	gtk_container_set_border_width(GTK_CONTAINER(_bin_dialog), 8);
	g_free(title);
	
	vbox = GTK_BOX(GTK_DIALOG(_bin_dialog)->vbox);
	gtk_box_set_spacing(vbox, 4);
	table = gtk_table_new(3, 2, FALSE);
	gtk_table_set_row_spacings(GTK_TABLE(table), 10);
	gtk_table_set_col_spacings(GTK_TABLE(table), 10);
	gtk_box_pack_start(vbox, table, FALSE, FALSE, 6);

	/* Name:  [______________] */
	label = gtk_label_new(_("Name:"));
	gtk_misc_set_alignment(GTK_MISC(label), 0.0, 0.0);
	gtk_table_attach(GTK_TABLE(table), label, 0, 1, 0, 1, GTK_FILL, 0, 0, 0);
	pacoball = g_strdup_printf("%s-paco.tar.gz", pkg_name_8);
	name_label = gtk_label_new(pacoball);
	gtk_misc_set_alignment(GTK_MISC(name_label), 0.0, 0.0);
	gtk_table_attach_defaults(GTK_TABLE(table), name_label, 1, 2, 0, 1);
	
	/* Save in folder:  [____________|v] */
	label = gtk_label_new(_("Save in folder:"));
	gtk_misc_set_alignment(GTK_MISC(label), 0.0, 0.0);
	gtk_table_attach(GTK_TABLE(table), label, 0, 1, 1, 2, GTK_FILL, 0, 0, 0);
	dir_button = gtk_file_chooser_button_new(NULL, GTK_FILE_CHOOSER_ACTION_SELECT_FOLDER);
	gtk_file_chooser_set_show_hidden(GTK_FILE_CHOOSER(dir_button), TRUE);
	gtk_file_chooser_set_current_folder(GTK_FILE_CHOOSER(dir_button), g_get_home_dir());
	gtk_table_attach_defaults(GTK_TABLE(table), dir_button, 1, 2, 1, 2);

	/* Compression:  [_gzip/bzip2________|v] */
	label = gtk_label_new(_("Compression:"));
	gtk_misc_set_alignment(GTK_MISC(label), 0.0, 0.0);
	gtk_table_attach(GTK_TABLE(table), label, 0, 1, 2, 3, GTK_FILL, 0, 0, 0);
	hbox = gtk_hbox_new(FALSE, 10);
	combo = gtk_combo_box_new_text();
	gtk_box_pack_start(GTK_BOX(hbox), combo, TRUE, TRUE, 0);
	gtk_combo_box_append_text(GTK_COMBO_BOX(combo), "gzip");
	gtk_combo_box_append_text(GTK_COMBO_BOX(combo), "bzip2");
	gtk_combo_box_set_active(GTK_COMBO_BOX(combo), GP_GZIP);
	g_signal_connect(combo, "changed", G_CALLBACK(zip_prog_change), (gpointer)name_label);
	
	/* Level: [_1____|v] */
	gtk_box_pack_start(GTK_BOX(hbox), gtk_label_new(_("Level:")), FALSE, FALSE, 2);
	combo = gtk_combo_box_new_text();
	gtk_box_pack_start(GTK_BOX(hbox), combo, FALSE, FALSE, 0);
	gtk_combo_box_append_text(GTK_COMBO_BOX(combo), _("1 (faster)"));
	for (i = 2; i < 9; i++) {
		sprintf(number, "%d" ,i);
		gtk_combo_box_append_text(GTK_COMBO_BOX(combo), number);
	}
	gtk_combo_box_append_text(GTK_COMBO_BOX(combo), _("9 (better)"));
	gtk_combo_box_set_active(GTK_COMBO_BOX(combo), 8);
	zip.level = 9;
	g_signal_connect(combo, "changed", G_CALLBACK(zip_level_change), NULL);
	
	gtk_table_attach_defaults(GTK_TABLE(table), hbox, 1, 2, 2, 3);
	
	hbox = GTK_DIALOG(_bin_dialog)->action_area;
	gtk_button_box_set_layout(GTK_BUTTON_BOX(hbox), GTK_BUTTONBOX_EDGE);

	g_signal_connect(_bin_dialog, "response", G_CALLBACK(bin_response), (gpointer)dir_button);
	g_signal_connect(_bin_dialog, "destroy", G_CALLBACK(bin_destroy), NULL);
	 
	gtk_widget_show_all(_bin_dialog);

	g_free(pacoball);
	g_free(pkg_name_8);
}


/* vim: set ts=4 */
