/*********************************************************************
 * files.c - The "Files" tab in the package window.
 *********************************************************************
 * This file is part of the package paco
 * Copyright (C) 2005 David Rosal <david.rosal@upf.edu>
 * For more information visit http://paco.sourceforge.net
 *********************************************************************/

#include "gpaco.h"
#include <gdk/gdkkeysyms.h>
#include <sys/stat.h>

#define ISGZIP(x) (g_str_has_suffix(x, ".gz") || g_str_has_suffix(x, ".tgz"))
#define ISBZIP2(x) g_str_has_suffix(x, ".bz2")

#define CHECK_TYPE(file, list_type)   \
	(file->miss ? (list_type & FILE_MISSING) : (list_type & FILE_INSTALLED))

enum {
	FILES_COL_NAME,
	FILES_COL_SIZE,
	FILES_COL_BG,
	FILES_COL_ID,
	FILES_COL_MISS,
	FILES_NUM_COLS
};


static GtkTreeModel *pkg_model_new(pkg_t *pkg)
{
	file_t *file;
	GtkTreeIter iter;
	GtkListStore *store;
	register GSList *f;
	register gchar *file_name_8;
	
	g_return_val_if_fail(pkg != NULL, NULL);
	
	store = gtk_list_store_new(FILES_NUM_COLS,
	  G_TYPE_STRING,	/* Name of the file */
	  G_TYPE_LONG,		/* Size in bytes */
	  G_TYPE_STRING,	/* "cell-background" attribute */
	  G_TYPE_LONG,		/* File id */
	  G_TYPE_BOOLEAN);	/* File missing */
	
	for (f = pkg->files; f; f = f->next) {
		file = (file_t *)(f->data);
		if (!CHECK_TYPE(file, pkg->list_type))
			continue;
		file_name_8 = gp_filename_to_utf8(file->name);
		gtk_list_store_append(store, &iter);
		gtk_list_store_set(store, &iter,
		  FILES_COL_NAME, file_name_8,
		  FILES_COL_SIZE, file->size,
		  FILES_COL_BG, file->miss ? "lightyellow" : NULL,
		  FILES_COL_ID, file->id,
		  FILES_COL_MISS, file->miss, -1);
		g_free(file_name_8);
	}
	
	gtk_tree_sortable_set_sort_func(GTK_TREE_SORTABLE(store), FILES_COL_NAME,
	 casesort, GINT_TO_POINTER(FILES_COL_NAME), NULL);
	
	return GTK_TREE_MODEL(store);
}


static file_t *get_file_from_path(pkg_t *pkg, GtkTreeModel *model, GtkTreePath *path)
{
	register GSList *f;
	file_t *file;
	GtkTreeIter iter;
	guint id;
	
	g_return_val_if_fail(pkg != NULL, NULL);
	
	gtk_tree_model_get_iter(model, &iter, path);
	gtk_tree_model_get(model, &iter, FILES_COL_ID, &id, -1);
	
	for (f = pkg->files; f; f = f->next) {
		file = (file_t *)(f->data);
		if (file->id == id)
			return file;
	}
	
	return NULL;
}


static GList *get_selected_files(pkg_t *pkg, gint list_type)
{
	file_t *file;
	GList *files = NULL, *s, *selected;
	GtkTreeModel *model;
	GtkTreePath *path;
	GtkTreeSelection *sel;
	
	g_return_val_if_fail(pkg != NULL, NULL);
	
	sel = gtk_tree_view_get_selection(GTK_TREE_VIEW(pkg->list_tv));
	selected = gtk_tree_selection_get_selected_rows(sel, &model);

	for (s = selected; s; s = s->next) {
		path = (GtkTreePath *)(s->data);
		file = get_file_from_path(pkg, model, path);
		if (CHECK_TYPE(file, list_type))
			files = g_list_append(files, file);
		gtk_tree_path_free(path);
	}

	g_list_free(selected);
	return files;
}


void files_treeview_update(pkg_t *pkg)
{
	GList *f, *files;
	GtkTreeSelection *sel;
	GtkTreeModel *model;
	GtkSortType order;
	GtkTreeIter iter;
	GtkTreeView *tv;
	gboolean got_scroll = FALSE, valid, sorted;
#if (GTK_MINOR_VERSION < 8)
	gint xw, yw;
	GdkRectangle r;
#endif
	gint sort_column_id;
	guint32 id, top_id;
	GtkTreePath *path;
	
	g_return_if_fail(pkg != NULL);
	g_return_if_fail(pkg->window != NULL);
	
	tv = (GTK_TREE_VIEW(pkg->list_tv));
	model = gtk_tree_view_get_model(tv);
	
	/* Get the scrolling status */
	path = gtk_tree_path_new();
#if (GTK_MINOR_VERSION >= 8)
	got_scroll = gtk_tree_view_get_visible_range(GTK_TREE_VIEW(pkg->list_tv), &path, NULL);
#else
	if (GTK_WIDGET_REALIZED(pkg->list_tv)) {
		gtk_tree_view_get_visible_rect(tv, &r);
		gtk_tree_view_tree_to_widget_coords(tv, r.x, r.y, &xw, &yw);
		got_scroll = gtk_tree_view_get_path_at_pos(tv, xw, yw, &path, NULL, NULL, NULL);
	}
#endif
	if (got_scroll && gtk_tree_model_get_iter(model, &iter, path))
		gtk_tree_model_get(model, &iter, FILES_COL_ID, &top_id, -1);
	gtk_tree_path_free(path);
	
	/* Get the "sorting status" of the tree view */
	sorted = gtk_tree_sortable_get_sort_column_id(GTK_TREE_SORTABLE(model),
	  &sort_column_id, &order);
	
	files = get_selected_files(pkg, FILE_LOGGED);
	
	/* Rebuild the model */
	gtk_tree_view_set_model(tv, NULL);
	model = pkg_model_new(pkg);

	/* Restore the sorting status */
	if (sorted)
		gtk_tree_sortable_set_sort_column_id(GTK_TREE_SORTABLE(model), sort_column_id, order);
	
	/* Connect the model to the tree view */
	gtk_tree_view_set_model(tv, model);
	g_object_unref(model);
	
	/* Restore the selection status */
	if (files) {
		sel = gtk_tree_view_get_selection(tv);
		valid = gtk_tree_model_get_iter_first(model, &iter);
		while (valid) {
			gtk_tree_model_get(model, &iter, FILES_COL_ID, &id, -1);
			gtk_tree_selection_unselect_iter(sel, &iter);
			for (f = files; f; f = f->next) {
				if (((file_t *)(f->data))->id == id) {
					gtk_tree_selection_select_iter(sel, &iter);
					break;
				}
			}
			valid = gtk_tree_model_iter_next(model, &iter);
		}
		g_list_free(files);
	}
	
	/* Restore the scrolling status */
	if (got_scroll) {
		valid = gtk_tree_model_get_iter_first(model, &iter);
		while (valid) {
			gtk_tree_model_get(model, &iter, FILES_COL_ID, &id, -1);
			if (id == top_id) {
				if ((path = gtk_tree_model_get_path(model, &iter))) {
					gtk_tree_view_scroll_to_cell(tv, path, NULL, TRUE, 0.0, 0.0);
					gtk_tree_path_free(path);
				}
				break;
			}
			valid = gtk_tree_model_iter_next(model, &iter);
		}
	}
	REFRESH_GUI;
}


static void switch_installed(__UNUSED GtkWidget *button, pkg_t *pkg)
{
	g_return_if_fail(pkg != NULL);
	
	pkg->list_type ^= FILE_INSTALLED; 
	
	if (pkg->n_inst)
		files_treeview_update(pkg);
}


static void switch_missing(__UNUSED GtkWidget *button, pkg_t *pkg)
{
	g_return_if_fail(pkg != NULL);
	
	pkg->list_type ^= FILE_MISSING;
	
	if (pkg->n_miss)
		files_treeview_update(pkg);
}


static void size_cell_data_func(__UNUSED GtkTreeViewColumn *col, GtkCellRenderer *cell,
                                GtkTreeModel *model, GtkTreeIter *iter, __UNUSED gpointer data)
{
	gint32 size;
	gchar *size_str;
	
	gtk_tree_model_get(model, iter, FILES_COL_SIZE, &size, -1);
	size_str = strsize((gfloat)size);
	g_object_set(cell, "text", size_str, "xalign", 1.0, "xpad", 10, NULL);
	g_free(size_str);
}


static guint32 get_files_size(GList *files)
{
	file_t *file;
	guint32 size;
	register GList *f;
	
	g_return_val_if_fail(files != NULL, 0);
	
	for (size = 0, f = files; f; f = f->next) {
		file = (file_t *)(f->data);
		if (file->size > 0)
			size += file->size;
	}

	return size;
}


static void remove_files(pkg_t *pkg)
{
	file_t *file;
	GList *files, *f;
	GtkWidget *pbar, *prog_dialog = NULL;
	gchar *question, *file_name_8;
	gint e, n, response, cnt;
	gfloat i;
	GString *errmsg;
	struct stat s;
	
	g_return_if_fail(pkg != NULL);
	g_return_if_fail((files = get_selected_files(pkg, FILE_INSTALLED)));
	
	if ((n = g_list_length(files)) == 1)
		question = g_strdup(_("Remove 1 file ?"));
	else
		question = g_strdup_printf(_("Remove %d files ?"), n);
	
	response = gp_question_dialog(pkg->window, question);
	g_free(question);
	if (response != GTK_RESPONSE_YES) {
		g_list_free(files);
		return;
	}
		
	if (get_files_size(files) > (4 * MEGABYTE) || pkg->n_inst > 1000)
		prog_dialog = gp_progress_dialog(_("Removing files..."), pkg->window, &pbar);
	
	errmsg = g_string_new("");
	
	for (e = cnt = 0, i = 0.0, f = files; f; f = f->next) {
		file = (file_t *)(f->data);
		if (prog_dialog) {
			gtk_progress_bar_set_fraction(GTK_PROGRESS_BAR(pbar), i++ / n);
			REFRESH_GUI;
		}
		if (g_lstat(file->name, &s) < 0 || g_unlink(file->name) == 0) {
			cnt++;
			file->miss = TRUE;
		}
		else if (++e < 16) {
			file_name_8 = gp_filename_to_utf8(file->name);
			g_string_append_printf(errmsg, " '%s': %s\n", file_name_8, g_strerror(errno));
			g_free(file_name_8);
		}
		else if (e == 16)
			g_string_append(errmsg, " ...");
	}

	REFRESH_GUI;
	
	if (errmsg->len) {
		gp_error_dialog(_("Error while removing files:"), pkg->window, errmsg->str);
		g_string_free(errmsg, TRUE);
	}
	
	if (cnt)
		update_pkg(pkg);

	g_list_free(files);
	
	if (prog_dialog)
		gtk_widget_destroy(prog_dialog);
}


static void invert_selection(pkg_t *pkg)
{
	GtkTreeSelection *sel;
	GtkTreeModel *model;
	GtkTreeIter iter;
	gboolean valid;

	g_return_if_fail(pkg != NULL);
	
	sel = gtk_tree_view_get_selection(GTK_TREE_VIEW(pkg->list_tv));
	model = gtk_tree_view_get_model(GTK_TREE_VIEW(pkg->list_tv));
	
	g_return_if_fail((valid = gtk_tree_model_get_iter_first(model, &iter)));
	
	do {
		if (gtk_tree_selection_iter_is_selected(sel, &iter))
			gtk_tree_selection_unselect_iter(sel, &iter);
		else
			gtk_tree_selection_select_iter(sel, &iter);
	}
	while ((valid = gtk_tree_model_iter_next(model, &iter)));
}


static gboolean write_files_label(pkg_t *pkg)
{
	GList *files;
	GtkWidget *label;
	guint32 size = 0, n_files = 0;
	gchar *s, *text, *size_str;

	g_return_val_if_fail(pkg != NULL, FALSE);
	
	if (!pkg->window)
		return FALSE;  /* stop the timeout func. */
		
	if (!(files = get_selected_files(pkg, pkg->list_type))) {
		if (pkg->list_type & FILE_INSTALLED) {
			n_files = pkg->n_inst;
			size = pkg->size_inst;
		}
		if (pkg->list_type & FILE_MISSING) {
			n_files += pkg->n_miss;
			size += pkg->size_miss;
		}
		s = g_strdup((n_files == 1) ? _("file") : _("files"));
	}
	else {
		n_files = g_list_length(files);
		s = g_strdup((n_files > 1) ? _("files selected") : _("file selected"));
		size = get_files_size(files);
		g_list_free(files);
	}
	
	size_str = strsize(size);
	text = g_strdup_printf("  %d %s  |  %s", n_files, s, size_str);
	
	label = g_object_get_data(G_OBJECT(pkg->window), "files-label");
	gtk_label_set_text(GTK_LABEL(label), text);

	g_free(size_str);
	g_free(s);
	g_free(text);
	
	return TRUE;
}


static void process(pkg_t *pkg, gint action)
{
	file_t *file;
	GList *f, *files;
	GtkWidget *dialog = NULL, *pbar;
	gchar prog[10], *msg = NULL, *error, *cmd;
	GString *errmsg;
	gint e, n;
	struct stat s;
	gfloat i;
	
	g_return_if_fail(pkg != NULL);
	g_return_if_fail((files = get_selected_files(pkg, FILE_INSTALLED)));
	n = g_list_length(files);

	if (get_files_size(files) > MEGABYTE) {
		if (action == GP_UNCOMPRESS)
			msg = g_strdup(_("Uncompressing files..."));
		else if (action == GP_GZIP)
			msg = g_strdup_printf(_("Compressing files with %s..."), "gzip");
		else if (action == GP_BZIP2)
			msg = g_strdup_printf(_("Compressing files with %s..."), "bzip2");
		dialog = gp_progress_dialog(msg, pkg->window, &pbar);
		g_free(msg);
	}
	
	errmsg = g_string_new("");
	
	for (e = 0, i = 0.0, f = files; f; f = f->next) {
		
		file = (file_t *)(f->data);
		
		if (dialog) {
			gtk_progress_bar_set_fraction(GTK_PROGRESS_BAR(pbar), i++ / n);
			REFRESH_GUI;
		}
		if (g_lstat(file->name, &s) < 0)
			continue;
		
		else if (action == GP_UNCOMPRESS) {
			if (ISGZIP(file->name))
				memcpy(prog, "gzip -d", 8);
			else if (ISBZIP2(file->name))
				memcpy(prog, "bzip2 -d", 9);
			else
				continue;
		}
		else if (action == GP_GZIP) { 
			if (!ISGZIP(file->name) && !ISBZIP2(file->name))
				memcpy(prog, "gzip", 5);
			else
				continue;
		}
		else if (action == GP_BZIP2) { 
			if (!ISGZIP(file->name) && !ISBZIP2(file->name))
				memcpy(prog, "bzip2", 6);
			else
				continue;
		}
		
		cmd = g_strdup_printf("%s %s", prog, file->name);
		if (!gp_spawn(&error, cmd)) {
			if (++e < 16)
				g_string_append(errmsg, error);
			else if (e == 16)
				g_string_append(errmsg, " ...");
		}
		g_free(cmd);
	}
	
	if (dialog) {
		gtk_progress_bar_set_fraction(GTK_PROGRESS_BAR(pbar), 1.0);
		gtk_widget_destroy(dialog);
	}
	
	if (errmsg->len) {
		gp_error_dialog((action == GP_UNCOMPRESS) ?
		  _("Error while uncompressing files:") : _("Error while compressing files:"),
		  pkg->window, errmsg->str);
	}
	
	update_pkg(pkg);
	g_list_free(files);
	g_string_free(errmsg, TRUE);
}


static void do_unzip(pkg_t *pkg)
{
	process(pkg, GP_UNCOMPRESS);
}


static void do_gzip(pkg_t *pkg)
{
	process(pkg, GP_GZIP);
}


static void do_bzip2(pkg_t *pkg)
{
	process(pkg, GP_BZIP2);
}


static gboolean button_press(GtkWidget *tv, GdkEventButton *event, pkg_t *pkg)
{
	file_t *file;
	GList *f, *files = NULL;
	GtkTreeSelection *sel;
	GtkTreeModel *model;
	GtkTreeIter it;
    GtkWidget *menu, *submenu, *item;
	gboolean anyfile, anysel = FALSE, anyzip = FALSE, anyunzip = FALSE;
	
	g_return_val_if_fail(pkg != NULL, FALSE);
	
	if (event->button != 3)
		return FALSE;
    
	sel = gtk_tree_view_get_selection(GTK_TREE_VIEW(tv));
	model = gtk_tree_view_get_model(GTK_TREE_VIEW(tv));
	
	if ((anyfile = gtk_tree_model_get_iter_first(model, &it))) {
		anysel = ((files = get_selected_files(pkg, FILE_INSTALLED)) != NULL);
		for (f = files; f && (!anyzip || !anyunzip); f = f->next) {
			file = (file_t *)(f->data);
			if (ISGZIP(file->name) || ISBZIP2(file->name))
			 	anyzip = TRUE;
			else
				anyunzip = TRUE;
		}
		g_list_free(files);
	}
	
	menu = gtk_menu_new();
	gp_menu_item(_("Select _all"), anyfile, menu, G_CALLBACK(gtk_tree_selection_select_all), sel);
	gp_menu_item(_("U_nselect all"), anyfile, menu,
	  G_CALLBACK(gtk_tree_selection_unselect_all), sel);
	gp_menu_item(_("_Invert selection"), anyfile, menu, G_CALLBACK(invert_selection), pkg);
	SEPARATOR_MENU_ITEM(menu);
	item = gp_menu_item(_("_Compress"), anyunzip, menu, NULL, NULL);
	gtk_menu_item_set_submenu(GTK_MENU_ITEM(item), submenu = gtk_menu_new());
	gp_menu_item("_Gzip", TRUE, submenu, G_CALLBACK(do_gzip), pkg);
	gp_menu_item("_Bzip2", TRUE, submenu, G_CALLBACK(do_bzip2), pkg);
	gp_menu_item(_("_Uncompress"), anyzip, menu, G_CALLBACK(do_unzip), pkg);
	SEPARATOR_MENU_ITEM(menu);
	gp_image_menu_item(GTK_STOCK_DELETE, _("_Remove..."), menu, anysel,
	  G_CALLBACK(remove_files), pkg);
	
	gtk_widget_show_all(menu);
    gtk_menu_popup(GTK_MENU(menu), NULL, NULL, NULL, NULL, (event) ?
	  event->button : 0, gdk_event_get_time((GdkEvent *)event));
	  
	return TRUE;
}


static gboolean key_press(__UNUSED GtkWidget *tv, GdkEventKey *event, pkg_t *pkg)
{
	if (event->type == GDK_KEY_PRESS && event->keyval == GDK_Delete) {
		remove_files(pkg);
		return TRUE;
	}
	return FALSE;
}


static void gp_files_column(const gchar *title, gint column_id,
                            GtkTreeCellDataFunc func, GtkWidget *treeview)
{
	GtkTreeViewColumn *col;
	GtkCellRenderer *cell;
	
	cell = gtk_cell_renderer_text_new();
	col = gtk_tree_view_column_new_with_attributes(title, cell, "text", column_id,
	  "cell-background", FILES_COL_BG, NULL);
	gtk_tree_view_column_set_resizable(col, TRUE);
	gtk_tree_view_column_set_sort_column_id(col, column_id);
	if (func) 
		gtk_tree_view_column_set_cell_data_func(col, cell, func, NULL, NULL);
	gtk_tree_view_append_column(GTK_TREE_VIEW(treeview), col);
}
	

GtkWidget *files_page(pkg_t *pkg)
{
	GtkTreeSelection *sel;
	GdkColor color;
	GtkWidget *but_installed, *but_missing, *but_update, *scroll, *vbox,
	  *butbox, *label, *labelbox;
	
	g_return_val_if_fail(pkg != NULL, NULL);
	
	pkg->list_type = FILE_INSTALLED;

	but_installed = gtk_check_button_new_with_mnemonic(_("View _installed"));
	gtk_tooltips_set_tip(_tips, but_installed,
	  _("List logged files that exist in the filesystem"), NULL);
	gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(but_installed), TRUE);
	g_signal_connect(but_installed, "toggled", G_CALLBACK(switch_installed), pkg);
	 
	but_missing = gtk_check_button_new_with_mnemonic(_("View _missing"));
	gtk_tooltips_set_tip(_tips, but_missing,
	  _("List logged files that were removed after installation"), NULL);
	gdk_color_parse("lightyellow", &color);
	gtk_widget_modify_bg(but_missing, GTK_STATE_ACTIVE, &color);
	gtk_widget_modify_base(but_missing, GTK_STATE_NORMAL, &color);
	gtk_widget_modify_base(but_missing, GTK_STATE_PRELIGHT, &color);
	g_signal_connect(but_missing, "toggled", G_CALLBACK(switch_missing), pkg);
	
	but_update = gp_button(GTK_STOCK_REFRESH, _("_Update"), NULL, NULL);
	gtk_tooltips_set_tip(_tips, but_update, _("Update the log of the package"), NULL);
	gtk_widget_set_sensitive(but_update, ISWRITABLE(_logdir));
	g_signal_connect_swapped(but_update, "clicked", G_CALLBACK(update_pkg), pkg);
	
	pkg->list_tv = gtk_tree_view_new_with_model(pkg_model_new(pkg));
	gp_files_column(_("File"), FILES_COL_NAME, NULL, pkg->list_tv);
	gp_files_column(_("Size"), FILES_COL_SIZE, size_cell_data_func, pkg->list_tv);
	gtk_tree_view_set_rules_hint(GTK_TREE_VIEW(pkg->list_tv), (_opt & SHOW_RULES));
	sel = gtk_tree_view_get_selection(GTK_TREE_VIEW(pkg->list_tv));
	gtk_tree_selection_set_mode(sel, GTK_SELECTION_MULTIPLE);
	scroll = gtk_scrolled_window_new(NULL, NULL);
    gtk_scrolled_window_set_policy(GTK_SCROLLED_WINDOW(scroll),
	  GTK_POLICY_AUTOMATIC, GTK_POLICY_AUTOMATIC);
	gtk_container_add(GTK_CONTAINER(scroll), pkg->list_tv);
	g_signal_connect(pkg->list_tv, "button-press-event", G_CALLBACK(button_press), pkg);
	g_signal_connect(pkg->list_tv, "key-press-event", G_CALLBACK(key_press), pkg);
	
	label = gtk_label_new(NULL);
	gtk_misc_set_alignment(GTK_MISC(label), 0.0, 0.5);
	g_object_set_data(G_OBJECT(pkg->window), "files-label", label);
	
	vbox = gtk_vbox_new(FALSE, 0);
	butbox = gtk_hbox_new(FALSE, 16);
	gtk_box_pack_start(GTK_BOX(butbox), but_installed, FALSE, FALSE, 0);
	gtk_box_pack_start(GTK_BOX(butbox), but_missing, FALSE, FALSE,0);
	gtk_box_pack_end(GTK_BOX(butbox), but_update, FALSE, FALSE, 0);
	gtk_box_pack_start(GTK_BOX(vbox), butbox, FALSE, FALSE, 0);
	gtk_box_pack_start(GTK_BOX(vbox), scroll, TRUE, TRUE, 0);
	gtk_box_pack_start(GTK_BOX(vbox), gtk_hseparator_new(), FALSE, FALSE, 0);
	labelbox = gtk_hbox_new(FALSE, 16);
	gtk_box_pack_start(GTK_BOX(labelbox), label, FALSE, FALSE, 0);
	gtk_box_pack_start(GTK_BOX(vbox), labelbox, FALSE, FALSE, 0);
	
	g_timeout_add(128, (GSourceFunc)write_files_label, pkg);
	
	return vbox;
}


/* vim: set ts=4 */
