/*********************************************************************
 * gpaco.c
 *********************************************************************
 * This file is part of the package paco
 * Copyright (C) 2005 David Rosal <david.rosal@upf.edu>
 * For more information visit http://paco.sourceforge.net
 *********************************************************************/

#include "gpaco.h"
#include <signal.h>


/* Globals declaration */
window_t _window = { NULL, 500, 500, 0, 0 };
GtkTreeModel *_model;
GtkWidget *_bin_dialog;
GSList *_pkgs;
GtkTooltips *_tips;
gchar *_logdir, *_tmp;
gboolean _updating;
gint _blksize = HUMAN_READABLE;
int _opt = SHOW_NAME | SHOW_SIZE | SHOW_MISSING_SIZE | SHOW_DATE | SHOW_TIPS
             | SHOW_DAY | SHOW_ICON;


/* Recognized variables stored in the gpaco configuration file */
struct {
	gchar *tag;  /* Name in configuration file */
	void *var;   /* Place to store the value */
	gint flag;   /* Related flag (if type == PACO_FLAGS) */
	GType type;  /* Type of variable */
} gvar[] = {
	{ "SHOW_NAME",            &_opt, SHOW_NAME,            PACO_FLAGS },
	{ "SHOW_SIZE",            &_opt, SHOW_SIZE,            PACO_FLAGS },
	{ "SHOW_MISSING_SIZE",    &_opt, SHOW_MISSING_SIZE,    PACO_FLAGS },
	{ "SHOW_DATE",            &_opt, SHOW_DATE,            PACO_FLAGS },
	{ "SHOW_FILES",           &_opt, SHOW_FILES,           PACO_FLAGS },
	{ "SHOW_MISSING_FILES",   &_opt, SHOW_MISSING_FILES,   PACO_FLAGS },
	{ "SHOW_RULES",           &_opt, SHOW_RULES,           PACO_FLAGS },
	{ "SHOW_TIPS",            &_opt, SHOW_TIPS,            PACO_FLAGS },
	{ "SHOW_DAY",             &_opt, SHOW_DAY,             PACO_FLAGS },
	{ "SHOW_HOUR",            &_opt, SHOW_HOUR,            PACO_FLAGS },
	{ "SHOW_ICON",            &_opt, SHOW_ICON,            PACO_FLAGS },
	{ "MISS_SIZE_AS_PERCENT", &_opt, MISS_SIZE_AS_PERCENT, PACO_FLAGS },
	{ "BLOCK_SIZE",           &_blksize, 0,             PACO_INT },
	{ "WINDOW_WIDTH",         &_window.width, 0,        PACO_INT },
	{ "WINDOW_HEIGHT",        &_window.height, 0,       PACO_INT },
	{ "WINDOW_XPOS",          &_window.x, 0,            PACO_INT },
	{ "WINDOW_YPOS",          &_window.y, 0,            PACO_INT },
	{ NULL, NULL, 0, 0 }
};


static void clear_tmp()
{
	g_unlink(_tmp);
}


__attribute__((__noreturn__))
static RETSIGTYPE sighand(int sig)
{
	if (sig == SIGSEGV)
		g_printerr(_("Segmentation fault\nPlease report this bug to <%s>"), PACKAGE_BUGREPORT);
	exit(EXIT_FAILURE);
}


__attribute__((__noreturn__, __format__(printf, 1, 2)))
static void die(const gchar *fmt, ...)
{
	va_list ap;
	gchar msg[KILOBYTE];

	va_start(ap, fmt);
	vsnprintf(msg, sizeof(msg), fmt, ap);
	va_end(ap);

	g_printerr(msg);
	
	exit(EXIT_FAILURE);
}


static void init_globals()
{
	gint fd, i, fatal[] = FATAL_SIGNALS;
	gchar *gpacorc, *pacorc;
	
	/* Fatal signal handlers */
	for (i = 0; fatal[i] != 0; i++)
		signal(fatal[i], sighand);
		
	pacorc = g_build_filename(CONFDIR, "pacorc", NULL);
	gpacorc = g_build_filename(g_get_home_dir(), ".gpacorc", NULL);
	
	paco_getconf(pacorc, "APPARENT_SIZE", &_opt, PACO_FLAGS, APPARENT_SIZE);

	if (!_logdir && !paco_getconf(pacorc, "LOGDIR", &_logdir, PACO_STRING, 0))
		_logdir = g_strdup(LOGDIR);
	
	/* Parse the gpaco configuration file and get the GUI options. */
	for (i = 0; gvar[i].tag; i++)
		paco_getconf(gpacorc, gvar[i].tag, gvar[i].var, gvar[i].type, gvar[i].flag);
	
	_tips = gtk_tooltips_new();
	if (!(_opt & SHOW_TIPS))
		gtk_tooltips_disable(_tips);

	if ((fd = g_file_open_tmp("gpacoXXXXXX", &_tmp, NULL)) < 0)
		die("g_file_open_tmp(): %s", g_strerror(errno));
	close(fd);
	g_atexit(clear_tmp);
	
	g_free(gpacorc);
	g_free(pacorc);
}


/* Save the GUI settings in gpacorc */
static void write_gpacorc()
{
	gint i;
	FILE *f;
	gchar *gpacorc;

	gpacorc = g_strdup_printf("%s/.gpacorc", g_get_home_dir());
	g_return_if_fail((f = g_fopen(gpacorc, "w")));
	
	for (i = 0; gvar[i].tag; i++) {
		if (gvar[i].type == PACO_FLAGS) {
			if (*(int *)gvar[i].var & gvar[i].flag)
				fprintf(f, "%s=1\n", gvar[i].tag);
			else
				fprintf(f, "%s=0\n", gvar[i].tag);
		}
		else if (gvar[i].type == PACO_INT)
			fprintf(f, "%s=%d\n", gvar[i].tag, *(gint *)gvar[i].var);
	}
	fclose(f);
	g_free(gpacorc);
}


static void error_func(__UNUSED GOptionContext *context, __UNUSED GOptionGroup *group,
                       __UNUSED gpointer data, GError **error)
{
	gchar *help;
	
	if (*error && (*error)->message) {
		help = g_strdup_printf(_("\nTry '%s --help' for more information"), "gpaco");
		die("%s%s", (*error)->message, help);
	}
}


static void parse_command_line(int argc, char *argv[])
{
	GOptionContext *context;
	GError *error = NULL;
	GOptionEntry opt[] = {
		{ "logdir", 'L', 0, G_OPTION_ARG_FILENAME, &_logdir,
		 N_("Use DIR as the log directory."), N_("DIR") },
		{ NULL, 0, 0, 0, NULL, NULL, NULL }
	};
	
	context = g_option_context_new(_("- GTK+ frontend of paco."));
	g_option_context_add_main_entries(context, opt, NULL);
	g_option_group_set_error_hook(g_option_context_get_main_group(context),
	  (GOptionErrorFunc)error_func);
	g_option_context_parse(context, &argc, &argv, &error);
	g_option_context_free(context);
}


static void printerr_handler(const gchar *msg)
{
	g_fprintf(stderr, "gpaco: %s\n", msg);
}


/* Read the logs and store the package information in a GSList. */
static GSList *get_list_of_pkgs()
{
	GSList *list = NULL;
	GDir *dir;
	G_CONST_RETURN gchar *entry;
	gchar *logdir_8;
	pkg_t *pkg;
	struct stat s;
	gint id = 0;

	logdir_8 = gp_filename_to_utf8(_logdir);
	
	if (paco_inpaths(_logdir, DEFAULT_EXCLUDE))
		die(_("%s: Invalid log directory"), logdir_8);
	else if (g_lstat(_logdir, &s) < 0)
		die("%s: %s", logdir_8, g_strerror(errno));
	else if (!S_ISDIR(s.st_mode))
		die(_("%s: Not a directory"), logdir_8);
	else if (!(dir = g_dir_open(_logdir, 0, NULL)))
		die(_("Can't open log directory '%s'"), logdir_8);
	
	while ((entry = g_dir_read_name(dir))) {
		if ((pkg = new_pkg((gchar *)entry, &id)))
			list = g_slist_prepend(list, (gpointer)pkg);
	}
	
	g_free(logdir_8);
	g_dir_close(dir);
	return list;
}


int main(int argc, char *argv[])
{
#ifdef ENABLE_NLS
	gchar *localedir;
#endif

	gtk_init(&argc, &argv);

#ifdef ENABLE_NLS
	localedir = g_build_filename(DATADIR, "locale", NULL);
	bindtextdomain("paco", localedir);
	g_free(localedir);
	bind_textdomain_codeset("paco", "UTF-8");
	textdomain("paco");
#endif

	g_set_printerr_handler((GPrintFunc)printerr_handler);
	
	init_globals();
	parse_command_line(argc, argv);
	
	g_atexit(write_gpacorc);
	
	_pkgs = get_list_of_pkgs();
	main_window_new();
	gtk_main();
	
	exit(EXIT_SUCCESS);
}


/* vim: set ts=4 */
