/*********************************************************************
 * pkglist.c - Package data storage with GSList's
 *********************************************************************
 * This file is part of the package paco
 * Copyright (C) 2004-2005 David Rosal <david.rosal@upf.edu>
 * For more information visit http://paco.sourceforge.net
 *********************************************************************/

#include "gpaco.h"


__attribute__((__noreturn__))
static void parse_error(gchar *file, FILE *f)
{
	GtkWidget *dialog;
	gchar *msg;
	
	fclose(f);
	
	msg = g_strdup_printf(_("Parse error while reading '%s'.\n"
	  "Run '%s' to rewrite the logs"), file, "paco -u");
	dialog = gtk_message_dialog_new(NULL, 0, GTK_MESSAGE_ERROR, GTK_BUTTONS_OK, msg);
	gtk_window_set_title(GTK_WINDOW(dialog), _("gpaco - error"));
	gtk_dialog_run(GTK_DIALOG(dialog));
	
	exit(EXIT_FAILURE);
}


inline static void new_file_miss(pkg_t *pkg, int size, gint32 *id, const char *path)
{
	register file_t *file;

	file = g_new(file_t, 1);
	file->name = g_strdup(path);
	file->id = (*id)++;
	file->miss = TRUE;
	if ((file->size = size) > 0)
		pkg->size_miss += (gfloat)size;

	pkg->n_miss++;
	pkg->files = g_slist_prepend(pkg->files, (gpointer)file);
}


__attribute__((__format__(printf, 4, 5)))
inline static void new_file_inst(pkg_t *pkg, int size, gint32 *id, const char *fmt, ...)
{
	register file_t *file;
	va_list a;
	char path[PATH_MAX];

	va_start(a, fmt);
	vsnprintf(path, sizeof(path), fmt, a);
	va_end(a);
	
	file = g_new(file_t, 1);
	file->name = g_strdup(path);
	file->id = (*id)++;
	file->miss = FALSE;
	
	if ((file->size = size) > 0)
		pkg->size_inst += (gfloat)size;
	
	pkg->n_inst++;
	pkg->files = g_slist_prepend(pkg->files, (gpointer)file);
}


static void pkg_file_free(file_t *file, __UNUSED gpointer data)
{
	g_free(file->name);
	file->name = NULL;
}


void free_pkg_data(pkg_t *pkg, __UNUSED gpointer data)
{
	g_slist_foreach(pkg->files, (GFunc)pkg_file_free, NULL);
	g_slist_free(pkg->files);
	g_free(pkg->log);
	g_free(pkg->info);
	g_free(pkg->name);
	g_free(pkg->icon);
	pkg->files = NULL;
	pkg->info = pkg->log = pkg->name = pkg->icon = NULL;
}


static gint compare_func(gconstpointer a, gconstpointer b)
{
	return g_ascii_strcasecmp(((file_t *)a)->name, ((file_t *)b)->name);
}


void get_pkg_files(pkg_t *pkg)
{
	gint n, raw, bz2, gz;
	gchar file[PATH_MAX];
	FILE *f;
	gint32 id = 0, size;

	g_return_if_fail(pkg != NULL);
	
	pkg->n_miss = pkg->n_inst = pkg->size_inst = pkg->size_miss = 0;
	
	if (pkg->files) {
		g_slist_foreach(pkg->files, (GFunc)pkg_file_free, NULL);
		g_slist_free(pkg->files);
		pkg->files = NULL;
	}
	
	/* Skip header */
	g_return_if_fail(pkg->log != NULL);
	g_return_if_fail((f = g_fopen(pkg->log, "r")));
	g_return_if_fail(fgets(file, sizeof(file), f));
	
	while ((n = fscanf(f, "%[^|]|%d|%d|%d\n", file, &raw, &gz, &bz2)) != EOF) {
	
		if (n != 4 || !memchr("/-", *file, 2)) {
			if (!_updating)
				parse_error(pkg->log, f);
			continue;
		}
		else if (*file == '-') {
			/* missing file */
			size = (raw > 0) ? raw : 0;
			size += (gz > 0) ? gz : 0;
			size += (bz2 > 0) ? bz2 : 0;
			new_file_miss(pkg, size, &id, file + 1);
		}
		else {
			if (raw != SIZE_MISSING)
				new_file_inst(pkg, raw, &id, "%s", file);
			if (gz != SIZE_MISSING)
				new_file_inst(pkg, gz, &id, "%s.gz", file);
			if (bz2 != SIZE_MISSING)
				new_file_inst(pkg, bz2, &id, "%s.bz2", file);
		}
	}
	
	fclose(f);
	pkg->files = g_slist_sort(pkg->files, (GCompareFunc)compare_func);
}


pkg_t *new_pkg(gchar *name, gint *id)
{
	pkg_t *pkg = NULL;
	FILE *f, *finfo;
	gchar *k, *logfile, buf[KILOBYTE];

	if (ispunct((int)*name))
		return NULL;

	/* Check for the '#!paco' header (security issue) */
	logfile = g_build_filename(_logdir, name, NULL);
	g_return_val_if_fail((f = g_fopen(logfile, "r")), NULL);
	fgets(buf, 64, f);
	fclose(f);
	g_return_val_if_fail(buf && !strncmp(buf, "#!paco", 6), NULL);
	
	pkg = g_new0(pkg_t, 1);

	pkg->name = g_strdup(name);
	
	if ((k = strchr(buf, ' ')))
		pkg->date = atoi(k);
	
	pkg->log = logfile;
	pkg->info = g_build_filename(_logdir, "_info", name, NULL);
	
	if ((finfo = g_fopen(pkg->info, "r"))) {
		while (paco_fgets(buf, sizeof(buf), finfo) && !pkg->icon) {
			if (!strncmp(buf, "#i:/", 4))
				pkg->icon = g_strdup(&buf[3]);
		}
		fclose(finfo);
	}
	else {
		g_free(pkg->info);
		pkg->info = NULL;
	}
	
	get_pkg_files(pkg);
	pkg->id = (*id)++;
	
	return pkg;
}


/* vim: set ts=4 */
