/*********************************************************************
 * util.c - Global utilities.
 *********************************************************************
 * This file is part of the package paco
 * Copyright (C) 2005 David Rosal <david.rosal@upf.edu>
 * For more information visit http://paco.sourceforge.net
 *********************************************************************/

#include "gpaco.h"

#if HAVE_SYS_WAIT_H
#	include <sys/wait.h>
#endif
#ifndef WEXITSTATUS
#	define WEXITSTATUS(stat_val) ((unsigned)(stat_val) >> 8)
#endif
#ifndef WIFEXITED
#	define WIFEXITED(stat_val) (((stat_val) & 255) == 0)
#endif


gboolean gp_spawn(gchar **error, gchar *cmd)
{
	int status;

	g_spawn_command_line_sync(cmd, NULL, error, &status, NULL);
	if (!WIFEXITED(status) || WEXITSTATUS(status) != EXIT_SUCCESS) {
		if (*error && **error)
			return FALSE;
	}
	return TRUE;
}


gchar *gp_filename_to_utf8(gchar *str)
{
	gchar *str_8;
	GError *error;
	
	g_return_val_if_fail(str != NULL, NULL);

	if (g_utf8_validate(str, -1, NULL))
		return g_strdup(str);
	
	else if (!(str_8 = g_filename_to_utf8(str, -1, NULL, NULL, &error))) {
		if (error && error->message)
			g_printerr("g_filename_to_utf8(\"%s\", ...): %s", str, error->message);
		return g_strdup(str);
	}
	
	return str_8;
}


gchar *gp_filename_from_utf8(gchar *str_8)
{
	gchar *str;
	GError *error = NULL;
	
	g_return_val_if_fail(str_8 != NULL, NULL);

	if (!g_utf8_validate(str_8, -1, NULL))
		return g_strdup(str_8);
	
	else if (!(str = g_filename_from_utf8(str_8, -1, NULL, NULL, &error))) {
		if (error)
			g_printerr("g_filename_from_utf8(\"%s\", ...): %s", str_8, error->message);
		return g_strdup(str_8);
	}
	
	return str;
}


gchar *gp_locale_to_utf8(gchar *str)
{
	gchar *str_8;
	GError *error = NULL;
	
	g_return_val_if_fail(str != NULL, NULL);

	if (g_utf8_validate(str, -1, NULL))
		return g_strdup(str);
	
	else if (!(str_8 = g_locale_to_utf8(str, -1, NULL, NULL, &error))) {
		if (error)
			g_printerr("g_locale_to_utf8(\"%s\", ...): %s", str, error->message);
		return g_strdup(str);
	}
	
	return str_8;
}


GtkWidget *gp_image_menu_item(const gchar *stock_id, gchar *text, GtkWidget *menu,
                              gboolean sensitive, GCallback cb, gpointer cb_data)
{
	GtkWidget *item, *image;
	GdkPixbuf *icon;
	
	if (!text)
	    item = gtk_image_menu_item_new_from_stock(stock_id, NULL);
	else {
		item = gtk_image_menu_item_new_with_mnemonic(text);
		icon = gtk_widget_render_icon(item, stock_id, GTK_ICON_SIZE_MENU, NULL);
		image = gtk_image_new_from_pixbuf(icon);
		gtk_image_menu_item_set_image(GTK_IMAGE_MENU_ITEM(item), image);
	}
	gtk_container_add(GTK_CONTAINER(menu), item);
	gtk_widget_set_sensitive(item, sensitive);
	if (cb)
		g_signal_connect_swapped(G_OBJECT(item), "activate", cb, cb_data);
	
	return item;
}


GtkWidget *gp_check_menu_item(const gchar *title, gboolean active,
                              GtkWidget *menu, GCallback cb, gpointer cb_data)
{
	GtkWidget *item;
	
	item = gtk_check_menu_item_new_with_mnemonic(title);
	gtk_check_menu_item_set_active(GTK_CHECK_MENU_ITEM(item), active);
	g_signal_connect_swapped(item, "toggled", cb, cb_data);
	gtk_container_add(GTK_CONTAINER(menu), item);

	return item;
}


GtkWidget *gp_menu(const gchar *label, GtkWidget *menubar)
{
	GtkWidget *menu, *item;
	
	menu = gtk_menu_new();
	item = gtk_menu_item_new_with_mnemonic(label);
	gtk_container_add(GTK_CONTAINER(menubar), item);
	gtk_menu_item_set_submenu(GTK_MENU_ITEM(item), menu);

	return menu;
}


GtkWidget *gp_menu_item(const gchar *title, gboolean sensitive, GtkWidget *menu,
                        GCallback cb, gpointer cb_data)
{
	GtkWidget *item;
	
	item = gtk_menu_item_new_with_mnemonic(title);
	gtk_widget_set_sensitive(item, sensitive);
	if (cb)
		g_signal_connect_swapped(item, "activate", cb, cb_data);
	gtk_container_add(GTK_CONTAINER(menu), item);
	
	return item;
}


void gp_column(const gchar *title, gint col_id, gboolean visible,
               gfloat xalign, GtkTreeCellDataFunc func, GtkWidget *treeview)
{
	GtkTreeViewColumn *col;
	GtkCellRenderer *cell;
	
	cell = gtk_cell_renderer_text_new();
	gtk_cell_renderer_set_fixed_size(cell, -1, 24);
	col = gtk_tree_view_column_new_with_attributes(title, cell, "text", col_id, NULL);
	if (func)
		gtk_tree_view_column_set_cell_data_func(col, cell, func, GINT_TO_POINTER(col_id), NULL);
	gtk_tree_view_column_set_resizable(col, TRUE);
	gtk_tree_view_column_set_sort_column_id(col, col_id);
	gtk_tree_view_column_set_visible(col, visible);
	gtk_tree_view_column_set_alignment(col, xalign);
	gtk_tree_view_column_set_min_width(col, 16);
	gtk_tree_view_append_column(GTK_TREE_VIEW(treeview), col);
}

	
GtkWidget *gp_button(const gchar *stock_id, const gchar *text, GCallback cb, gpointer cb_data)
{
	GtkWidget *hbox, *button, *image, *label;
	GdkPixbuf *icon;
	
	hbox = gtk_hbox_new(FALSE, 5);
	button = gtk_button_new();
	icon = gtk_widget_render_icon(button, stock_id, GTK_ICON_SIZE_BUTTON, NULL);
	image = gtk_image_new_from_pixbuf(icon);
	label = gtk_label_new_with_mnemonic(text);
	if (cb)
		g_signal_connect_swapped(G_OBJECT(button), "clicked", cb, cb_data);
	gtk_box_pack_start(GTK_BOX(hbox), image, FALSE, FALSE, 0);
	gtk_box_pack_start(GTK_BOX(hbox), label, FALSE, FALSE, 0);
	gtk_container_add(GTK_CONTAINER(button), hbox);
	
	return button;
}


gint gp_question_dialog(GtkWidget *parent, gchar *txt)
{
	GtkWidget *dialog;
	gint response;

	dialog = gtk_message_dialog_new(GTK_WINDOW(parent), GTK_DIALOG_DESTROY_WITH_PARENT,
	  GTK_MESSAGE_QUESTION, GTK_BUTTONS_YES_NO, txt);
	response = gtk_dialog_run(GTK_DIALOG(dialog));
	gtk_widget_destroy(dialog);

	return response;
}

	
void gp_error_dialog(gchar *msg, GtkWidget *parent, gchar *errmsg)
{
	GtkWidget *dialog, *label;
	
	dialog = gtk_message_dialog_new(GTK_WINDOW(parent), GTK_DIALOG_DESTROY_WITH_PARENT,
	 GTK_MESSAGE_ERROR, GTK_BUTTONS_OK, msg);
	gtk_window_set_default_size(GTK_WINDOW(dialog), 300, -1);
	gtk_window_set_resizable(GTK_WINDOW(dialog), TRUE);
	label = gtk_label_new(errmsg);
	gtk_label_set_ellipsize(GTK_LABEL(label), PANGO_ELLIPSIZE_MIDDLE);
	gtk_box_pack_start(GTK_BOX(GTK_DIALOG(dialog)->vbox), label, FALSE,FALSE,0);
	gtk_widget_show_all(dialog);
	gtk_dialog_run(GTK_DIALOG(dialog));
	gtk_widget_destroy(dialog);
}


gint casesort(GtkTreeModel *model, GtkTreeIter *a, GtkTreeIter *b, gpointer p)
{
	gchar *name_a, *name_b;
	gint ret;
	
	gtk_tree_model_get(model, a, GPOINTER_TO_INT(p), &name_a, -1);
	gtk_tree_model_get(model, b, GPOINTER_TO_INT(p), &name_b, -1);
	ret = g_ascii_strcasecmp(name_a, name_b);
	g_free(name_a);
	g_free(name_b);
	
	return ret;
}


void gp_unlink(const gchar *path)
{
	if (path && g_unlink(path) < 0)
		g_printerr("unlink(\"%s\"): %s", path, g_strerror(errno));
}


/* Size printing. */
gchar *strsize(gfloat size)
{
	gchar units[2] = "";
	
	if (size == SIZE_SYMLINK)
		return g_strdup("@");
	else if (_blksize == HUMAN_READABLE || _blksize == 0) {
		if (size < KILOBYTE)
			return g_strdup_printf("%d", (gint)size);
		else if (size < MEGABYTE)
			return g_strdup_printf("%dk", (gint)(size / KILOBYTE));
		else if (size < 10 * MEGABYTE)
			return g_strdup_printf("%.1fM", size / MEGABYTE);
		else
			return g_strdup_printf("%dM", (gint)(size / MEGABYTE));
	}
	else if (_blksize == KILOBYTE)
		strcpy(units, "k");
	else if (_blksize == MEGABYTE)
		strcpy(units, "M");

	return g_strdup_printf("%d%s", (gint)(size / _blksize), units);
}


GtkWidget *gp_progress_dialog(char *msg, GtkWidget *window, GtkWidget **pbar)
{
	GtkWidget *dialog, *label;

	dialog = gtk_dialog_new();
	gtk_container_set_border_width(GTK_CONTAINER(dialog), 8);
	gtk_dialog_set_has_separator(GTK_DIALOG(dialog), FALSE);
	label = gtk_label_new(g_strdup_printf("    %s    \n", msg));
	gtk_container_add(GTK_CONTAINER(GTK_DIALOG(dialog)->vbox), label);
	if (pbar) {
		*pbar = gtk_progress_bar_new();
		gtk_container_add(GTK_CONTAINER(GTK_DIALOG(dialog)->vbox), *pbar);
	}
	if (window)
		gtk_window_set_transient_for(GTK_WINDOW(dialog), GTK_WINDOW(window));
	gtk_window_set_resizable(GTK_WINDOW(dialog), FALSE);
	gtk_window_set_modal(GTK_WINDOW(dialog), TRUE);
	g_signal_connect(dialog, "delete-event", G_CALLBACK(gtk_true), NULL);
	gtk_widget_show_all(dialog);
	REFRESH_GUI;
	
	return dialog;
}


/* vim: set ts=4 */
