/********************************************************************
 * paco.c - The main source file.
 ********************************************************************
 * This file is part of the package paco
 * Copyright (C) 2004-2005 David Rosal i Ricart <david.rosal@upf.edu>
 * For more information visit http://paco.sourceforge.net
 ********************************************************************/

#include "paco.h"
#include <sys/stat.h>
#include <getopt.h>
#include <signal.h>

#ifdef ENABLE_NLS
#	include <locale.h>
#endif

enum {
	MODE_PKGS = 0,
	MODE_FILES,
	MODE_UPDATE,
	MODE_UNLOG,
	MODE_QUERY,
	MODE_INFO,
	MODE_VAR,
	MODE_LOG,
	MODE_REMOVE,
	MODE_CONF,
	N_MODES
};

enum {
	OPT_SORT = 1,
	OPT_REMOVE_BATCH,
	OPT_REMOVE_SHARED,
	OPT_VERSION,
	OPT_IGNORE_ERRORS
};

static log_opts_t op_log;
static list_opts_t op_list;
static rm_opts_t op_rm;
static int all_pkgs;
static char **env_var, mode[N_MODES];


/* Globals declaration */
int _exit_status = EXIT_SUCCESS, _expand, _verbosity, _apparent;
char *_pacorc, *_logdir, _tmp[PATH_MAX];


__attribute__((__noreturn__))
static void help()
{
printf(_(
".o0(paco)0o. the source code pacKAGE oRGANIZER\n\n"
"Usage:\n"
"  paco [OPTIONS] <packages|files|command>\n\n"));
printf(_(
"General options:\n"
"  -L, --logdir=DIR         Use DIR as the log directory.\n"
"  -x, --expand             Expand the command line package names.\n"
"  -a, --all                Apply to all logged packages (not with -r).\n"
"  -v, --verbose            Verbose output (-vv produces debugging messages).\n"
"  -h, --help               Display this help message.\n"
"      --version            Display version information.\n\n"));
printf(_(
"Database maintenance options:\n"
"  -u, --update             Update the log of the package.\n"
"  -U, --unlog              Remove the log of the package. See Remove options."
"\n\n"));
printf(_(
"General list options:\n"
"  -b, --block-size=SIZE    Use blocks of SIZE bytes for the sizes.\n"
"  -k, --kilobytes          Like '--block-size=1024'.\n"
"      --sort=WORD          Sort by WORD: 'name', 'date', 'size', 'files',\n"
"                           'missing-size' or 'missing-files'.\n"
"  -R, --reverse            Reverse order while sorting.\n"
"  -t, --total              Print totals.\n\n"));
printf(_(
"Package list options:\n"
"  -1, --one-column         List one package per line.\n"
"  -F                       Print the number of installed files.\n"
"  -M                       Print the number of missing files.\n"
"  -C                       Print the number of shared files.\n"
"  -d, --date               Show installation date (-dd show the hour too).\n"
"  -s, --size               Show the installed size of each package.\n"
"  -n, --missing-size       Print the total missing size of the package.\n\n"));
printf(_(
"File list options:\n"
"  -f, --files              List installed files.\n"
"  -m, --missing-files      List missing files.\n"
"  -c, --shared             List only the shared files (see the man page).\n"
"  -w, --who-shares         With -c: Print the packages that share each file.\n"
"  -y, --symlinks           Print the contents of symbolic links.\n"
"  -z, --no-package-name    Don't print the name of the package.\n"
"  -s, --size               Show the size of each file.\n\n"));
printf(_(	
"Information options:\n"
"  Note: Information may be not available for all packages.\n"
"  -i, --info               Print package information.\n"
"  -o, --configure-options  Print the options passed to configure when the\n"
"                           package was installed.\n"
"  -q, --query              Query for the packages that own one or more files."
"\n"
"  -V, --variable=VAR       Print the package build-time value of the\n"
"                           environment variable VAR (see the man page).\n\n"));
printf(_(
"Remove options:\n"
"  -r, --remove             Remove the (non shared) files of the package.\n"
"      --remove-shared      Remove also the shared files.\n"
"      --batch              Don't ask for confirmation when removing.\n"
"  -U, --unlog              Always remove the log of the package.\n"
"  -e, --skip=DIR:...       Don't remove files in these directories.\n\n"));
printf(_(
"Log options:\n"
"  -l, --log                Enable log mode. See the man page.\n"
"  -p, --package=PKG        Name of the package to log.\n" 
"  -D, --dirname            Use the name of the current directory as the name\n"
"                           of the package.\n"
"  -+, --append             With -p or -D: If the package is already logged,\n"
"                           append the list of files into the log.\n"
"      --ignore-errors      Do not exit if the monitored command fails.\n"
"  -I, --include=DIR:...    List of paths to scan.\n"
"  -E, --exclude=DIR:...    List of paths to skip.\n\n"));
printf(_(
"Note: The package list mode is enabled by default.\n"
"\nSend bugs to: David Rosal i Ricart <%s>\n"), PACKAGE_BUGREPORT);

	exit(EXIT_SUCCESS);
}


__attribute__((__noreturn__))
static void version()
{
	printf("paco-%s  (%s)\nCopyright (C) David Rosal i Ricart <%s>\n%s\n",
	  PACO_VERSION, RELEASEDATE, PACKAGE_BUGREPORT,
	  _("Protected by the GNU General Public License"));
	
	exit(EXIT_SUCCESS);
}


static char *get_dirname()
{
	char dirname[PATH_MAX];

	if (!getcwd(dirname, sizeof(dirname)))
		die_errno("getcwd(%p, %d)", dirname, (int)sizeof(dirname));

	return strdup(strrchr(dirname, '/') + 1);
}


static void set_mode(int m, char c)
{
	size_t i, cnt;
	char aux[2] = "-", optstr[3] = "";
	
	mode[m] = c;
	
	for (i = cnt = 0; i < N_MODES; i++) {
		if (mode[i]) {
			aux[0] = mode[i];
			strcat(optstr, aux);
			if (cnt++)
				die_help(_("'-%s': Incompatible options"), optstr);
		}
	}
}


/* Process the '--sort=WORD' option */
static int get_sorting(char *arg)
{
	char valid[KILOBYTE];
	int n = strlen(arg);
	
	if (!n)
		goto Fail;
	else if (!strncmp(arg, "size", n))
		return SORT_SIZE;
	else if (!strncmp(arg, "date", n))
		return SORT_DATE;
	else if (!strncmp(arg, "missing-size", n) && n > 8)
		return SORT_MSIZE;
	else if (!strncmp(arg, "missing-files", n) && n > 8)
		return SORT_N_MFILES;
	else if (!strncmp(arg, "files", n))
		return SORT_N_FILES;
	else if (strncmp(arg, "name", n)) {
		Fail:
		strncpy(valid,
		  "  - 'name'\n"
		  "  - 'size'\n"
		  "  - 'date'\n"
		  "  - 'files'\n"
		  "  - 'missing-files'\n"
		  "  - 'missing-size'", sizeof(valid));
		die(_("Invalid argument '%s' for option '--sort'.\n"
		 "Valid arguments are:\n%s"), arg, valid);
	}
	return SORT_NAME;
}


/* Process the '--block-size=SIZE' option */
static int get_blksize(char *arg)
{
	int i = -1, b = 0;
	
	while (isdigit((int)arg[++i]));
	switch (arg[i]) {
		case 'k': case 'K': b = KILOBYTE; break;
		case 'm': case 'M': b = MEGABYTE; break;
		case 'b': case 'B': case 0: b = 1; break;
		default: die(_("%s: Invalid block size"), arg);
	}
	return i ? (atoi(arg) * b) : b;
}


static char **parse_command_line(int argc, char **argv)
{
	int n_args, op, i;
	char *p, **args = NULL;
	struct stat st;
	const char *opstring = "1+ab:cCdDe:E:fFhiI:klL:mMnop:qrRstuUvV:wxyz";
	struct option opt[] = {
		/* General options */
		{ "help", 0, 0, 'h' },
		{ "version", 0, 0, OPT_VERSION },
		{ "logdir",	1, 0, 'L' },
		{ "verbose", 0, 0, 'v' },
		{ "expand", 0, 0, 'x' },
		{ "all", 0, 0, 'a' },
 		/* Database maintenance options */
		{ "update", 0, 0, 'u' },
		{ "unlog", 0, 0, 'U' },
	 	/* List options */
		{ "date", 0, 0, 'd' },
		{ "sort", 1, 0, OPT_SORT },
		{ "block-size", 1, 0, 'b' },
		{ "kilobytes", 0, 0, 'k' },
		{ "size", 0, 0, 's' },
		{ "missing-size", 0, 0, 'n' },
		{ "files", 0, 0, 'f' },
		{ "missing-files", 0, 0, 'm' },
		{ "shared", 0, 0, 'c' },
		{ "who-shares", 0, 0, 'w' },
		{ "reverse", 0, 0, 'R' },
		{ "total", 0, 0, 't' },
		{ "one-column", 0, 0, '1' },
		{ "symlinks", 0, 0, 'y' },
		{ "no-package-name", 0, 0, 'z' },
		/* Info options */
		{ "info", 0, 0, 'i' },
		{ "query", 0, 0, 'q' },
		{ "configure-options", 0, 0, 'o' },
		{ "variable", 1, 0, 'V' },
		/* Remove options */
		{ "remove", 0, 0, 'r' },
		{ "remove-shared", 0, 0, OPT_REMOVE_SHARED },
		{ "batch", 0, 0, OPT_REMOVE_BATCH },
		{ "skip", 1, 0, 'e' },
		/* Log options */
		{ "log", 0, 0, 'l' },
		{ "package", 1, 0, 'p' },
		{ "include", 1, 0, 'I' },
		{ "exclude", 1, 0, 'E' },
		{ "append", 0, 0, '+' },
		{ "dirname", 0, 0, 'D' },
		{ "ignore-errors", 0, 0, OPT_IGNORE_ERRORS },
		{ NULL ,0, 0, 0 },
	};

		
	if (argc == 1)
		die_help(_("No input packages"));
	
	/* Deal with the weird non getopt-friendly '-p+' option */
    for (i = 1; i < argc; i++) {
        if (argv[i][0] == '-' && argv[i][1] != '-') {
			for (p = argv[i]; (p = strstr(p, "p+")); )
				memcpy(p, "+p", 2);
		}
    }
	
	/* Parse the command line */
    while ((op = getopt_long(argc, argv, opstring, opt, 0)) >= 0) {
        switch (op) {
			
			/* General options */
			case OPT_VERSION: version();
			case 'h': help();
			case 'L': _logdir = strdup(optarg); break;
			case 'v': _verbosity++; break;
			case 'x': _expand = 1; break;
			case 'a': all_pkgs = 1; break;
 			
			/* Database maintenance options */
			case 'u': set_mode(MODE_UPDATE, 'u'); break;
			case 'U':
				if (mode[MODE_REMOVE])
					op_rm.flags |= RM_UNLOG;
				else
					set_mode(MODE_UNLOG, 'U');
				break;
	 		
			/* General list options */
			case OPT_SORT: op_list.sort = get_sorting(optarg); break;
			case 'R': op_list.flags |= LIST_REVERSE; break;
			case 't': op_list.flags |= LIST_TOTAL; break;
			case 's': op_list.blksize = HUMAN_READABLE; break;
			case 'b': op_list.blksize = get_blksize(optarg); break;
			case 'k': op_list.blksize = KILOBYTE; break;
			
			/* Package list options */
			case 'd':
				set_mode(MODE_PKGS, 'd');
				op_list.p_flags |= (op_list.p_flags & PKG_DATE) ?
				  PKG_HOUR : PKG_DATE;
				break;
			case 'n':
				set_mode(MODE_PKGS, 'n');
				op_list.p_flags |= PKG_SIZE_MISS;
				break;
			case 'F':
				set_mode(MODE_PKGS, 'F');
				op_list.p_flags |= PKG_N_INST;
				break;
			case 'M':
				set_mode(MODE_PKGS, 'M');
				op_list.p_flags |= PKG_N_MISS;
				break;
			case 'C':
				set_mode(MODE_PKGS, 'C');
				op_list.p_flags |= PKG_N_SHARED;
				break;
			case '1': op_list.p_flags |= PKG_ONE_COLUMN; break;
			
			/* File list options */
			case 'f':
				set_mode(MODE_FILES, 'f');
				op_list.f_type |= FILE_INSTALLED;
				break;
			case 'm':
				set_mode(MODE_FILES, 'm');
				op_list.f_type |= FILE_MISSING;
				break;
			case 'c':
				set_mode(MODE_FILES, 'c');
				op_list.f_type |= FILE_SHARED;
				break;
			case 'y': op_list.f_flags |= FILE_SYMLINKS; break;
			case 'z': op_list.f_flags |= FILE_NO_PKGNAME; break;
			case 'w': op_list.f_flags |= FILE_SHARER; break;
			
			/* Info options */
			case 'i': set_mode(MODE_INFO, 'i'); break;
			case 'o': set_mode(MODE_CONF, 'o'); break;
			case 'q': set_mode(MODE_QUERY, 'q'); break;
			case 'V':
				set_mode(MODE_VAR, 'V');
				realloc_strv(&env_var, optarg);
				break;
				
			/* Remove options */
			case 'r':
				if (mode[MODE_UNLOG]) {
					mode[MODE_UNLOG] = 0;
					op_rm.flags |= RM_UNLOG;
				}
				set_mode(MODE_REMOVE, 'r');
				break;
			case OPT_REMOVE_SHARED: op_rm.flags |= RM_SHARED; break;
			case OPT_REMOVE_BATCH: op_rm.flags |= RM_BATCH; break;
			case 'e': op_rm.skip = strdup(optarg); break;
			
			/* Log options */
			case 'l': set_mode(MODE_LOG, 'l'); break;
			case 'p': op_log.pkg = strdup(optarg); break;
			case 'I': op_log.include = strdup(optarg); break;
			case 'E': op_log.exclude = strdup(optarg); break;
			case '+': op_log.append = 1; break;
			case 'D': op_log.pkg = get_dirname(); break;
			case OPT_IGNORE_ERRORS: op_log.ignore_errors = 1; break;
			
			default: die_help(NULL);
		}
	}

	/* The non-option arguments */
	if ((n_args = argc - optind)) {
    	args = xmalloc0((n_args + 1) * sizeof(char *));
		for (i = 0; i < n_args; i++)
        	args[i] = strdup(argv[optind + i]);
	}
	else if (mode[MODE_QUERY])
		die_help(_("No input files"));
	else if ((!all_pkgs && !mode[MODE_LOG]) || mode[MODE_REMOVE])
		die_help(_("No input packages"));
	
	if (op_list.blksize) {
		op_list.p_flags |= PKG_SIZE_INST;
		op_list.f_flags |= FILE_SIZE;
	}
	else
		op_list.blksize = HUMAN_READABLE;
		
	if (op_log.pkg && (!isalnum((int)op_log.pkg[0])))
		die(_("'%s': Invalid package name"), op_log.pkg);
	
	if (op_list.f_type == FILE_SHARED)
		op_list.f_type |= FILE_LOGGED;
		
	/* log directory */
	if (!_logdir)
		paco_getconf(_pacorc, "LOGDIR", (void *)&_logdir, PACO_STRING, 0);
	if (!_logdir)
		_logdir = strdup(LOGDIR);
	if (paco_inpaths(_logdir, DEFAULT_EXCLUDE))
		die(_("%s: Invalid log directory"), _logdir);
	if (!lstat(_logdir, &st) && !S_ISDIR(st.st_mode))
		die(_("%s: Not a directory"), _logdir);
	
	if (all_pkgs && !mode[MODE_LOG] && !mode[MODE_QUERY]) {
	 	xfree2(args);
		return ALL_PACKAGES;
	}
	else
		return mode[MODE_LOG] ? args : unics(args);
}


/* Check whether the database is reasonably new. */
static void checkdb()
{
	struct stat st;
	char *stamp;
	int maxdbage = 7;  /* default is one week  */
	
	paco_getconf(_pacorc, "MAX_DB_AGE", (void *)&maxdbage, PACO_INT, 0);
	if (maxdbage == 0)
		return;  /* age unlimited */
	
	xasprintf(&stamp, "%s/.updated", _logdir);
	
	if (lstat(stamp, &st) < 0)
		paco_touch(stamp);
	else if ((time(0) - st.st_mtime) > (maxdbage * 24 * 3600)) {
		fprintf(stderr, _("*** Warning: Database is more than %d days old.\n"
 		 "*** Run 'paco -ua' to update it.\n"), maxdbage);
	}
	
	free(stamp);
}


__attribute__((__noreturn__))
static RETSIGTYPE sighand(int sig)
{
	if (sig == SIGSEGV) {
		die(_("Segmentation fault\nPlease report this bug to <%s>"),
		 PACKAGE_BUGREPORT);
	}
	exit(EXIT_FAILURE);
}


static void tmp_clean()
{
	unlink(_tmp);
}


static void init_globals()
{
	int i, fatal[] = FATAL_SIGNALS;
	
#if ENABLE_NLS
	char *localedir;
	
	xasprintf(&localedir, "%s/locale", DATADIR);
	setlocale(LC_ALL, "");
	bindtextdomain("paco", localedir);
	free(localedir);
	textdomain("paco");
#endif

	/* Fatal signal handlers */
	for (i = 0; fatal[i]; i++)
		signal(fatal[i], sighand);

	/* Global temp. file name */
 	fclose(tmp_fopen(_tmp, sizeof(_tmp)));
	atexit(tmp_clean);

	xasprintf(&_pacorc, "%s/pacorc", CONFDIR);
	
	paco_getconf(_pacorc, "APPARENT_SIZE", (void *)&_apparent, PACO_INT, 0);
}


int main(int argc, char *argv[])
{
	char **args;

	init_globals();
	args = parse_command_line(argc, argv);

	if (mode[MODE_FILES])
		do_list_files(args, &op_list);
	else if (mode[MODE_REMOVE])
		do_remove(args, &op_rm);
	else if (mode[MODE_LOG])
		do_log(args, &op_log);
	else if (mode[MODE_UNLOG])
		do_unlog(args);
	else if (mode[MODE_UPDATE])
		do_update(args);
	else if (mode[MODE_QUERY])
		do_query(args);
	else if (mode[MODE_INFO])
		do_info(args);
	else if (mode[MODE_CONF])
		do_conf(args);
	else if (mode[MODE_VAR])
		do_var(args, env_var);
	else
		do_list_pkgs(args, &op_list);
	
	checkdb();
	exit(_exit_status);
}


/* vim: set ts=4: */
