/*********************************************************************
 * pkglist.c - Managing package information with singly linked lists
 *             (called list_t's).
 *********************************************************************
 * This file is part of the package paco
 * Copyright (C) 2004-2005 David Rosal i Ricart <david.rosal@upf.edu>
 * For more information visit http://paco.sourceforge.net
 *********************************************************************/

#include "paco.h"
#include <stdarg.h>


int list_length(list_t *list)
{
	int ret = 0;
	list_t *li;
	
	for (li = list; li; li = li->next)
		ret++;
	
	return ret;
}


list_t *list_reverse(list_t *list)
{
	list_t *prev = NULL;
	
	while (list) {
		list_t *next = list->next;
		list->next = prev;
		prev = list;
		list = next;
	}
	return prev;
}


static list_t *list_prepend(list_t *list, void *data)
{
	list_t *new = xmalloc(sizeof(list_t));
	new->data = data;
	new->next = list;
	return new;
}


void list_foreach(list_t *list, void (*func)(void *, void *), void *data)
{
	while (list) {
		list_t *next = list->next;
		(*func)(list->data, data);
		list = next;
	}
}


static void free_file(void *data, __UNUSED void *user_data)
{
	file_t *file;
	
	if ((file = (file_t *)data)) {
		xfree(file->name);
		free(file);
	}
}


static void free_pkg(void *data, __UNUSED void *user_data)
{
	pkg_t *pkg;
	
	if ((pkg = (pkg_t *)data)) {
		list_foreach(pkg->files, free_file, NULL);
		xfree(pkg->files);
		xfree(pkg->log);
		xfree(pkg->info);
		xfree(pkg->name);
		free(pkg);
	}
}


void pkg_list_free(list_t *list)
{
	if (list) {
		list_foreach(list, free_pkg, NULL);
		free(list);
	}
}


static list_t *list_merge(list_t *p, list_t *q, list_cmp_f *cmp)
{
	list_t *r, head;

	for (r = &head; p && q; ) {
		if (cmp(p, q) > 0) {
			r = r->next = p;
			p = p->next;
		}
		else {
			r = r->next = q;
			q = q->next;
		}
	}
	r->next = p ? p : q;
	return head.next;
}


list_t *list_sort(list_t *p, list_cmp_f *cmp)
{
	list_t *q, *r;

	if (!(q = p))
		return NULL;
	if (!(r = p->next))
		return p;
	
	for ( ; r && (r = r->next) != NULL; r = r->next)
		q = q->next;
	
	r = q->next;
	q->next = NULL;
	
	if (r)
		p = list_merge(list_sort(r, cmp), list_sort(p, cmp), cmp);
		
	return p;
}


/* Check whether the string 'str' is a valid reference to the 'pkg'.
 * Example: given the pkg 'vorbis-tools-1.0', the string 'vorbis-tools'
 * match it, but not 'vorbis' or 'vorbis-tools-2.0'.
 * This provides a kind of automatic pkg name completion for the command
 * line arguments, whose targets are the names of the logged pkgs.
 */
static int match_pkg(char *str, char *pkgname)
{
	size_t len = strlen(str);
	char *dot, *dash;

	if (!_expand)
		return !strcasecmp(pkgname, str);
	else if (strncasecmp(pkgname, str, len))
		return 0;

	dot = strrchr(str, '.');
	dash = strrchr(str, '-');
	
	if (pkgname[len] == '\0'
	 || ( dot && (isdigit((int)*(dot + 1)) || isdigit((int)*(dot - 1))) )
	 || ( dash && isdigit((int)*(dash + 1)) )
	 || ( ispunct((int)pkgname[len]) && isdigit((int)pkgname[len + 1]) ))
	 	return 1;
	else
		return 0;
}


int sort_files_by_name(list_t *p, list_t *q)
{
	return strcasecmp(((file_t *)q->data)->name, ((file_t *)p->data)->name);
}


int sort_files_by_size(list_t *p, list_t *q)
{
	return ((file_t *)p->data)->size - ((file_t *)q->data)->size;
}


void sort_pkg_files_by_name(void *data, __UNUSED void *user_data)
{
	pkg_t *pkg = (pkg_t *)data;
	pkg->files = list_sort(pkg->files, sort_files_by_name);
}


static void new_file_miss(pkg_t *pkg, int size, const char *path)
{
	register file_t *file;

	file = xmalloc(sizeof(file_t));
	file->name = strdup(path);
	file->shared = 0;
	
	if ((file->size = size) > 0)
		pkg->size_miss += size;

	pkg->n_miss++;
	pkg->files = list_prepend(pkg->files, file);
}


__attribute__((__format__(printf, 3, 4)))
static void new_file_inst(pkg_t *pkg, int size, const char *fmt, ...)
{
	register file_t *file;
	va_list a;
	char path[PATH_MAX];

	va_start(a, fmt);
	vsnprintf(path, sizeof(path), fmt, a);
	va_end(a);
	
	file = xmalloc(sizeof(file_t));
	file->name = strdup(path);
	file->shared = 0;
	
	if ((file->size = size) > 0)
		pkg->size_inst += size;
	
	pkg->n_inst++;
	pkg->files = list_prepend(pkg->files, file);
}


#define PARSE_ERROR(file, log)  do { \
	fclose(file); \
	die(_("Parse error while reading '%s'.\n" \
	 "Run 'paco -ua' to rewrite the logs"), log); \
} while (0)


/* Read a log and get information of the files */
static void get_pkg_files(void *data, void *user_data)
{
	char file[PATH_MAX], buf[PATH_MAX];
	FILE *f;
	int type, n;
	long raw, bz2, gz, size;
	pkg_t *pkg;
	
	pkg = (pkg_t *)data;
	type = *(int *)user_data;

	pkg->n_miss = pkg->n_inst = pkg->size_inst = pkg->size_miss = 0;
	
	/* Skip header */
	f = paco_fopen(pkg->log, "r");
	fgets(buf, sizeof(buf), f);
	
	/* Read the list of files */
	while (paco_fgets(buf, sizeof(buf), f)) {
		
		if ((n = sscanf(buf, "%[^|]|%ld|%ld|%ld", file, &raw, &gz, &bz2)) != 4)
			PARSE_ERROR(f, pkg->log);
			
		else if (*file == '/') {
			if (type & FILE_INSTALLED) {
				if (raw != SIZE_MISSING)
					new_file_inst(pkg, raw, "%s", file);
				if (gz != SIZE_MISSING)
					new_file_inst(pkg, gz, "%s.gz", file);
				if (bz2 != SIZE_MISSING)
					new_file_inst(pkg, bz2, "%s.bz2", file);
			}
		}
		
		else if (!memcmp(file, "-/", 2)) {
			/* Missing file */
			if (type & FILE_MISSING) {
				size = (raw > 0) ? raw : 0;
				size += (gz > 0) ? gz : 0;
				size += (bz2 > 0) ? bz2 : 0;
				new_file_miss(pkg, size, file + 1);
			}
		}
		
		else
			PARSE_ERROR(f, pkg->log);
	}
	
	fclose(f);
}


static void get_pkg_totals(void *data, void *user_data)
{
	pkg_t *pkg;
	total_t *total;

	pkg = (pkg_t *)data;
	total = (total_t *)user_data;

	total->n_inst += pkg->n_inst;
	total->n_miss += pkg->n_miss;
	total->size_inst += pkg->size_inst;
	total->size_miss += pkg->size_miss;
}


void get_totals(list_t *pkgs, total_t *total)
{
	list_foreach(pkgs, get_pkg_totals, (void *)total);
}


void get_files(list_t *pkgs, int type)
{
	list_foreach(pkgs, get_pkg_files, (void *)&type);
}


static pkg_t *new_pkg(char *name)
{
	FILE *f;
	char *logfile, buf[64];
	pkg_t *pkg;
	
	if (ispunct((int)*name))
		return NULL;
		
	xasprintf(&logfile, "%s/%s", _logdir, name);
	if (!(f = fopen(logfile, "r"))) {
		free(logfile);
		return NULL;
	}
	
	/* Check for the '#!paco' header (security issue) */
	if (!fgets(buf, sizeof(buf), f) || strncmp(buf, "#!paco", 6)) {
		free(logfile);
		fclose(f);
		return NULL;
	}

	fclose(f);
	
	pkg = xmalloc0(sizeof(pkg_t));
	pkg->log = logfile;
	pkg->name = strdup(name);
	pkg->date = xatoi(strchr(buf, ' '));
	
	xasprintf(&pkg->info, "%s/_info/%s", _logdir, name);
	if (access(pkg->info, R_OK) < 0) {
		free(pkg->info);
		pkg->info = NULL;
	}
	
	return pkg;
}


/* Find all logged pkgs that match any name in 'match', and store
 * them in a list_t. */
static list_t *get_matching_pkgs(char *match[])
{
	list_t *pkgs = NULL;
	pkg_t *pkg;
	struct dirent *d;
	int logged, i;
	DIR *dir;
		
	if (!(dir = opendir(_logdir)))
		die_errno("opendir(\"%s\")", _logdir);
		
	for (i = 0; match[i]; i++) {
		for (logged = 0; (d = readdir(dir)); ) {
			if (match_pkg(match[i], d->d_name) && (pkg = new_pkg(d->d_name))) {
				logged++;
				pkgs = list_prepend(pkgs, (void *)pkg);
			}
		}
		rewinddir(dir);
		if (!logged) {
			_exit_status = EXIT_FAILURE;
			say(_("paco: %s: package not logged.\n"), match[i]);
		}
	}
	
	closedir(dir);
	return pkgs;
}


/* Store all logged pkgs in a list_t. */
static list_t *get_all_pkgs()
{
	list_t *pkgs = NULL;
	struct dirent *d;
	DIR *dir;
	pkg_t *pkg;
	
	if (!(dir = opendir(_logdir)))
		die_errno("opendir(\"%s\")", _logdir);
	
	while ((d = readdir(dir))) {
		if ((pkg = new_pkg(d->d_name)))
			pkgs = list_prepend(pkgs, (void *)pkg);
	}
	closedir(dir);
	
	return pkgs;
}


/* Read the logs and store the pkg information in a list_t. */
list_t *get_pkgs(char *match[])
{
	list_t *pkgs;

	if (match == ALL_PACKAGES)
		pkgs = get_all_pkgs();
	else
		pkgs = get_matching_pkgs(match);
	
	if (!pkgs) {
		if (match == ALL_PACKAGES)
			say(_("paco: no packages logged in directory '%s'\n"), _logdir);
		exit(EXIT_FAILURE);
	}
	
	return pkgs;
}


/* vim: set ts=4: */
