/*********************************************************************
 * remove.c - Removing and unlogging packages
 *********************************************************************
 * This file is part of the package paco
 * Copyright (C) 2004-2005 David Rosal i Ricart <david.rosal@upf.edu>
 * For more information visit http://paco.sourceforge.net
 *********************************************************************/

#include "paco.h"
#include <sys/stat.h>

static void removedir(char *dir);


static int isshared(char *my_file, char *my_pkg, list_t *all_pkgs)
{
	list_t *p, *f;
	file_t *file;
	pkg_t *pkg;
	int n;
	
	for (p = all_pkgs; p; p = p->next) {
		pkg = (pkg_t *)(p->data);
		if (!strcmp(pkg->name, my_pkg))
			continue;  /* skip myself */
		for (f = pkg->files; f; f = f->next) {
			file = (file_t *)(f->data);
			if (!(n = strcmp(file->name, my_file)))
				return 1;
			else if (n > 0)
				break;
		}
	}
	return 0;
}


static void remove_parent_dir(char *path)
{
	char *s, *p, *dir;
	
	dir = strdup(path);
	p = dir + strlen(dir) - 1;
	
	while (p != dir && *p == '/')
		*p-- = '\0';
	
	if ((s = strrchr(dir + 1, '/'))) {
		*s = '\0';
		if (strchr(dir + 1, '/'))
			removedir(dir);
	}
	free(dir);
}


static void removedir(char *dir)
{
	if (!rmdir(dir)) {
		say(_("Removed directory '%s'\n"), dir);
		remove_parent_dir(dir);
	}
	else if (errno != ENOTEMPTY)
		say_errno("rmdir(\"%s\")", dir);
}


static void remove_log(char *path)
{
	if (!path)
		return;
	else if (!unlink(path))
		say(_("Removed log '%s'\n"), path);
	else
		say_errno("unlink(\"%s\")", path);
}


static void unlog(void *data, __UNUSED void *user_data)
{
	remove_log(((pkg_t *)data)->log);
	remove_log(((pkg_t *)data)->info);
}


static int removefile(file_t *file)
{
	if (unlink(file->name) < 0) {
		say_errno("unlink(\"%s\")", file->name);
		return 0;
	}
	else if (file->size == SIZE_SYMLINK)
		say(_("Removed symlink '%s'\n"), file->name);
	else
		say(_("Removed '%s'\n"), file->name);
	
	remove_parent_dir(file->name);
	
	return 1;
}


static int readyes()
{
	char answer[KILOBYTE << 3], *p;
	
	if (!paco_fgets(answer, sizeof(answer), stdin))
		return 0;
	
	for (p = answer; isspace((int)*p); p++);
	
	return (!strcmp(p, "y") || !strcmp(p, "yes"));
}


/* size printing */
static char *size_str(long size)
{
	static char str[64];
	
	if (size < KILOBYTE)
		snprintf(str, sizeof(str), "%ld", size);
	else if (size < MEGABYTE)
		snprintf(str, sizeof(str), "%ldk", size / KILOBYTE);
	else if (size < (10 * MEGABYTE))
		snprintf(str, sizeof(str), "%.1fM", (float)size/MEGABYTE);
	else
		snprintf(str, sizeof(str), "%ldM", size / MEGABYTE);
		
	return str;
}


static void remove_pkg(void *data, void *user_data)
{
	list_t *f, *all_pkgs;
	file_t *file;
	struct stat s;
	int keep_logs = 0, removed_files = 0;
	rm_opts_t *opt = (rm_opts_t *)user_data;
	pkg_t *pkg = (pkg_t *)data;
	unsigned long removed_size = 0;
	
	all_pkgs = get_pkgs(ALL_PACKAGES);
	get_files(all_pkgs, FILE_LOGGED);

	if (!(opt->flags & RM_SHARED))
		list_foreach(all_pkgs, sort_pkg_files_by_name, NULL);
		
	/* If needed, ask for confirmation */
	if (!(opt->flags & RM_BATCH)) {
		printf("Remove package '%s' (y/n) ? ", pkg->name);
		if (!readyes())
			return;
	}
	
	/* Remove files */
	for (f = pkg->files; f; f = f->next) {
		file = (file_t *)(f->data);
		if (lstat(file->name, &s) < 0)
			continue;
		else if (paco_inpaths(file->name, opt->skip)) {
			say(_("%s: skipped\n"), file->name);
			keep_logs++;
		}
		else if (!(opt->flags & RM_SHARED)
		  && isshared(file->name, pkg->name, all_pkgs)) {
			say(_("%s: shared\n"), file->name);
			keep_logs++;
		}
		else if (!removefile(file))
			keep_logs++;
		else if (_verbosity >= PACO_DEBUG) {
			removed_size += s.st_size;
			removed_files++;
		}
	}
	
	pkg_list_free(all_pkgs);
	
	/* Remove the logs? */
	if (!keep_logs || opt->flags & RM_UNLOG) {
		if (!(opt->flags & RM_BATCH)) {
			printf("Remove the logs of the package %s (y/n) ? ", pkg->name);
			if (!readyes())
				goto Return;
		}
		unlog(pkg, NULL);
	}
	
	Return:
	debug(0, _("%d files removed (%s)\n"), removed_files,
	  size_str(removed_size));
	  
}


/* Main remove function */
void do_remove(char *args[], rm_opts_t *opt)
{
	list_t *pkgs;
	
	pkgs = get_pkgs(args);
	get_files(pkgs, FILE_LOGGED);
	
	list_foreach(pkgs, remove_pkg, (void *)opt);
	
	pkg_list_free(pkgs);
}


/* Just remove the log(s) of the pkg(s) */
void do_unlog(char *args[])
{
	list_t *pkgs;
	
	pkgs = get_pkgs(args);
	list_foreach(pkgs, unlog, NULL);	
	pkg_list_free(pkgs);
}


/* vim: set ts=4: */
